import argparse
import logging
import asyncio
import sys
import json

logger = logging.getLogger(__name__)


def cli(args=None):
    parser = argparse.ArgumentParser(description="jupyterhub client cli")
    subparser = parser.add_subparsers(help="jupyterhub client cli")
    create_run_subcommand(subparser)
    create_token_subcommand(subparser)
    parser.set_defaults(func=None)
    parser.add_argument(
        "-v", "--verbose", action="store_true", help="turn on jhub_client debugging"
    )
    args = parser.parse_args(args)

    if args.func is None:
        parser.print_help(sys.stderr)
        sys.exit(1)

    logging_level = logging.DEBUG if args.verbose else logging.INFO
    logging.basicConfig(level=logging_level)
    args.func(args)


def create_token_subcommand(subparser):
    subparser = subparser.add_parser("token")
    subparser.add_argument(
        "--hub",
        type=str,
        default="http://localhost:8000",
        help="url for running jupyterhub cluster",
    )
    subparser.add_argument(
        "--name",
        default="jhub-client",
        type=str,
        help="name to give to jupyterhub api token",
    )
    subparser.add_argument(
        "--no-verify-ssl",
        action="store_true",
        help="disable ssl checking",
    )
    subparser.set_defaults(func=handle_token)


def handle_token(args):
    from jhub_client.api import JupyterHubAPI

    async def create_token(hub_url):
        async with JupyterHubAPI(
            hub_url, auth_type="basic", verify_ssl=not args.no_verify_ssl
        ) as hub:
            print(hub.api_token, end="")

    loop = asyncio.get_event_loop()
    loop.run_until_complete(create_token(hub_url=args.hub))


def create_run_subcommand(subparser):
    subparser = subparser.add_parser("run")
    subparser.add_argument(
        "-n", "--notebook", type=str, help="notebook to run", required=True
    )
    subparser.add_argument(
        "--auth-type",
        type=str,
        default="token",
        choices=["token", "basic", "keycloak"],
        help="jupyterhub authentication type to use with default of token based",
    )
    subparser.add_argument(
        "--hub",
        type=str,
        default="http://localhost:8000",
        help="url for running jupyterhub cluster with default of 'http://localhost:8000'",
    )
    subparser.add_argument(
        "-u", "--username", type=str, help="username to run notebook as"
    )
    subparser.add_argument(
        "--user-options", type=str, help="json object representing user server options"
    )
    subparser.add_argument(
        "--temporary-user",
        action="store_true",
        default=False,
        help="create user temporarily if does not exist",
    )
    subparser.add_argument(
        "-d",
        "--daemonize",
        action="store_true",
        default=False,
        help="run notebook asyncronously",
    )
    subparser.add_argument(
        "--stop-server",
        action="store_true",
        default=False,
        help="stop server after completion of notebook",
    )
    subparser.add_argument(
        "--validate",
        action="store_true",
        default=False,
        help="validate notebook output matches",
    )
    subparser.add_argument(
        "--kernel-spec",
        type=str,
        help="kernel spec to launch is not specified will use default",
    )
    subparser.add_argument(
        "--output-filename",
        type=str,
        help="output filename for results of running notebook",
    )
    subparser.add_argument(
        "--no-verify-ssl",
        action="store_true",
        help="disable ssl checking",
    )
    subparser.set_defaults(func=handle_run)


def handle_run(args):
    from jhub_client.execute import execute_notebook
    from jhub_client.utils import render_notebook

    loop = asyncio.get_event_loop()

    try:
        user_options = json.loads(args.user_options or "{}")
    except json.decoder.JSONDecodeError:
        logger.error(f'unable to json parse user options="{args.user_options}"')
        sys.exit(1)

    kwargs = {
        "hub_url": args.hub,
        "notebook_path": args.notebook,
        "username": args.username,
        "temporary_user": args.temporary_user,
        "create_user": args.temporary_user,
        "delete_user": args.temporary_user,
        "validate": args.validate,
        "daemonized": args.daemonize,
        "stop_server": args.stop_server,
        "user_options": user_options,
        "kernel_spec": args.kernel_spec,
        "auth_type": args.auth_type,
        "verify_ssl": not args.no_verify_ssl,
    }

    if args.daemonize and args.temporary_user:
        logger.warning(
            "running notebook in daemonized mode will not delete temporary user"
        )

    if args.daemonize and args.validate:
        logger.error("running notebook in daemonized mode does not support validation")
        sys.exit(1)

    if args.daemonize and args.output_filename:
        logger.error(
            "running notebooks in daemonized mode does not support writing output to notebook"
        )
        sys.exit(1)

    cell_results = loop.run_until_complete(execute_notebook(**kwargs))

    if args.output_filename:
        output_notebook = render_notebook(cell_results)

        with open(args.output_filename, "w") as f:
            json.dump(output_notebook, f, indent=4)



import os
import uuid
import logging
import time
import asyncio

import yarl
import aiohttp

from jhub_client import auth


logger = logging.getLogger(__name__)


class JupyterHubAPI:
    def __init__(self, hub_url, auth_type="token", verify_ssl=True, **kwargs):
        self.hub_url = yarl.URL(hub_url)
        self.api_url = self.hub_url / "hub/api"
        self.auth_type = auth_type
        self.verify_ssl = verify_ssl

        if auth_type == "token":
            # Only look for the env var if api_token isn't explicitly passed.
            # Otherwise, we always try to read the env var even if it isn't necessary,
            # and this can crash if the env var isn't set.
            self.api_token = (
                kwargs.get("api_token") or os.environ["JUPYTERHUB_API_TOKEN"]
            )
        elif auth_type == "basic" or auth_type == "keycloak":
            self.username = kwargs.get("username") or os.environ["JUPYTERHUB_USERNAME"]
            self.password = kwargs.get("password") or os.environ["JUPYTERHUB_PASSWORD"]

    async def __aenter__(self):
        if self.auth_type == "token":
            self.session = await auth.token_authentication(
                self.api_token, verify_ssl=self.verify_ssl
            )
        elif self.auth_type == "basic":
            self.session = await auth.basic_authentication(
                self.hub_url, self.username, self.password, verify_ssl=self.verify_ssl
            )
            self.api_token = await self.create_token(self.username)
            await self.session.close()
            logger.debug("upgrading basic authentication to token authentication")
            self.session = await auth.token_authentication(
                self.api_token, verify_ssl=self.verify_ssl
            )
        elif self.auth_type == "keycloak":
            self.session = await auth.keycloak_authentication(
                self.hub_url, self.username, self.password, verify_ssl=self.verify_ssl
            )
            self.api_token = await self.create_token(self.username)
            await self.session.close()
            logger.debug("upgrading keycloak authentication to token authentication")
            self.session = await auth.token_authentication(
                self.api_token, verify_ssl=self.verify_ssl
            )
        return self

    async def __aexit__(self, exc_type, exc, tb):
        await self.session.close()

    async def ensure_user(self, username, create_user=False):
        user = await self.get_user(username)
        if user is None:
            if create_user:
                await self.create_user(username)
            else:
                raise ValueError(
                    f"current username={username} does not exist and create_user={create_user}"
                )
            user = await self.get_user(username)
        return user

    async def get_user(self, username):
        async with self.session.get(self.api_url / "users" / username) as response:
            if response.status == 200:
                return await response.json()
            elif response.status == 404:
                logger.info(f"username={username} does not exist")
                return None

    async def create_user(self, username):
        async with self.session.post(self.api_url / "users" / username) as response:
            if response.status == 201:
                logger.info(f"created username={username}")
                response = await response.json()
                self.api_token = await self.create_token(username)
                return response
            elif response.status == 409:
                raise ValueError(f"username={username} already exists")

    async def delete_user(self, username):
        async with self.session.delete(self.api_url / "users" / username) as response:
            if response.status == 204:
                logger.info(f"deleted username={username}")
            elif response.status == 404:
                raise ValueError(f"username={username} does not exist cannot delete")

    async def ensure_server(
        self, username, timeout, user_options=None, create_user=False
    ):
        user = await self.ensure_user(username, create_user=create_user)
        if user["server"] is None:
            await self.create_server(username, user_options=user_options)

        start_time = time.time()
        while True:
            user = await self.get_user(username)
            if user["server"] and user["pending"] is None:
                return JupyterAPI(
                    self.hub_url / "user" / username,
                    self.api_token,
                    verify_ssl=self.verify_ssl,
                )

            await asyncio.sleep(5)
            total_time = time.time() - start_time
            if total_time > timeout:
                logger.error(f"jupyterhub server creation timeout={timeout:.0f} [s]")
                raise TimeoutError(
                    f"jupyterhub server creation timeout={timeout:.0f} [s]"
                )

            logger.info(f"pending spawn polling for seconds={total_time:.0f} [s]")

    async def ensure_server_deleted(self, username, timeout):
        user = await self.get_user(username)
        if user is None:
            return  # user doesn't exist so server can't exist

        start_time = time.time()
        while True:
            server_status = await self.delete_server(username)
            if server_status == 204:
                return

            await asyncio.sleep(5)
            total_time = time.time() - start_time
            if total_time > timeout:
                logger.error(f"jupyterhub server deletion timeout={timeout:.0f} [s]")
                raise TimeoutError(
                    f"jupyterhub server deletion timeout={timeout:.0f} [s]"
                )

            logger.info(f"pending deletion polling for seconds={total_time:.0f} [s]")

    async def create_token(self, username, token_name=None):
        token_name = token_name or "jhub-client"
        async with self.session.post(
            self.api_url / "users" / username / "tokens", json={"note": token_name}
        ) as response:
            logger.info(f"created token for username={username}")
            return (await response.json())["token"]

    async def create_server(self, username, user_options=None):
        user_options = user_options or {}
        async with self.session.post(
            self.api_url / "users" / username / "server", json=user_options
        ) as response:
            logger.info(
                f"creating cluster username={username} user_options={user_options}"
            )
            if response.status == 400:
                raise ValueError(f"server for username={username} is already running")
            elif response.status == 201:
                logger.info(
                    f"created server for username={username} with user_options={user_options}"
                )
                return True

    async def delete_server(self, username):
        response = await self.session.delete(
            self.api_url / "users" / username / "server"
        )
        logger.info(f"deleted server for username={username}")
        return response.status

    async def info(self):
        async with self.session.get(self.api_url / "info") as response:
            return await response.json()

    async def list_users(self):
        async with self.session.get(self.api_url / "users") as response:
            return await response.json()

    async def list_proxy(self):
        async with self.session.get(self.api_url / "proxy") as response:
            return await response.json()

    async def identify_token(self, token):
        async with self.session.get(
            self.api_url / "authorizations" / "token" / token
        ) as response:
            return await response.json()


class JupyterAPI:
    def __init__(self, notebook_url, api_token, verify_ssl=True):
        self.api_url = yarl.URL(notebook_url) / "api"
        self.api_token = api_token or os.environ["JUPYTERHUB_API_TOKEN"]
        self.verify_ssl = verify_ssl

    async def __aenter__(self):
        self.session = aiohttp.ClientSession(
            headers={"Authorization": f"token {self.api_token}"},
            connector=aiohttp.TCPConnector(ssl=None if self.verify_ssl else False),
        )
        return self

    async def __aexit__(self, exc_type, exc, tb):
        await self.session.close()

    async def create_kernel(self, kernel_spec=None):
        data = {"kernel_spec": kernel_spec} if kernel_spec else None

        async with self.session.post(self.api_url / "kernels", json=data) as response:
            data = await response.json()
            logger.info(
                f'created kernel_spec={kernel_spec} kernel={data["id"]} for jupyter'
            )
            return data

    async def list_kernel_specs(self):
        async with self.session.get(self.api_url / "kernelspecs") as response:
            return await response.json()

    async def list_kernels(self):
        async with self.session.get(self.api_url / "kernels") as response:
            return await response.json()

    async def ensure_kernel(self, kernel_spec=None):
        kernel_specs = await self.list_kernel_specs()
        if kernel_spec is None:
            kernel_spec = kernel_specs["default"]
        else:
            available_kernel_specs = list(kernel_specs["kernelspecs"].keys())
            if kernel_spec not in kernel_specs["kernelspecs"]:
                logger.error(
                    f"kernel_spec={kernel_spec} not listed in available kernel specifications={available_kernel_specs}"
                )
                raise ValueError(
                    f"kernel_spec={kernel_spec} not listed in available kernel specifications={available_kernel_specs}"
                )

        kernel_id = (await self.create_kernel(kernel_spec=kernel_spec))["id"]
        return kernel_id, JupyterKernelAPI(
            self.api_url / "kernels" / kernel_id,
            self.api_token,
            verify_ssl=self.verify_ssl,
        )

    async def get_kernel(self, kernel_id):
        async with self.session.get(self.api_url / "kernels" / kernel_id) as response:
            if response.status == 404:
                return None
            elif response.status == 200:
                return await response.json()

    async def delete_kernel(self, kernel_id):
        async with self.session.delete(
            self.api_url / "kernels" / kernel_id
        ) as response:
            if response.status == 404:
                raise ValueError(
                    f"failed to delete kernel_id={kernel_id} does not exist"
                )
            elif response.status == 204:
                logger.info(f"deleted kernel={kernel_id} for jupyter")
                return True


class JupyterKernelAPI:
    def __init__(self, kernel_url, api_token, verify_ssl=True):
        self.api_url = kernel_url
        self.api_token = api_token or os.environ["JUPYTERHUB_API_TOKEN"]
        self.verify_ssl = verify_ssl

    async def __aenter__(self):
        self.session = aiohttp.ClientSession(
            headers={"Authorization": f"token {self.api_token}"},
            connector=aiohttp.TCPConnector(ssl=None if self.verify_ssl else False),
        )
        self.websocket = await self.session.ws_connect(self.api_url / "channels")
        return self

    async def __aexit__(self, exc_type, exc, tb):
        await self.session.close()

    def request_execute_code(self, msg_id, username, code):
        return {
            "header": {
                "msg_id": msg_id,
                "username": username,
                "msg_type": "execute_request",
                "version": "5.2",
            },
            "metadata": {},
            "content": {
                "code": code,
                "silent": False,
                "store_history": True,
                "user_expressions": {},
                "allow_stdin": True,
                "stop_on_error": True,
            },
            "buffers": [],
            "parent_header": {},
            "channel": "shell",
        }

    async def send_code(self, username, code, wait=True, timeout=None):
        msg_id = str(uuid.uuid4())

        await self.websocket.send_json(
            self.request_execute_code(msg_id, username, code)
        )

        if not wait:
            return None

        async for msg_text in self.websocket:
            if msg_text.type != aiohttp.WSMsgType.TEXT:
                return False

            # TODO: timeout is ignored

            msg = msg_text.json()

            if "parent_header" in msg and msg["parent_header"].get("msg_id") == msg_id:
                # These are responses to our request
                if msg["channel"] == "iopub":
                    if msg["msg_type"] == "execute_result":
                        return msg["content"]["data"]["text/plain"]
                    elif msg["msg_type"] == "stream":
                        return msg["content"]["text"]
                    # cell did not produce output
                    elif msg["content"].get("execution_state") == "idle":
                        return ""



import asyncio

from jhub_client.execute import execute_code


async def simulate_users(hub_url, num_users, user_generator, workflow="concurrent"):
    jupyterhub_sessions = []

    if workflow == "concurrent":
        for i, (username, cells) in zip(range(num_users), user_generator):
            jupyterhub_sessions.append(
                execute_code(
                    hub_url=hub_url,
                    username=username,
                    cells=cells,
                    create_user=True,
                    delete_user=True,
                )
            )

        return await asyncio.gather(*jupyterhub_sessions)
    else:
        raise ValueError("uknown type of jupyterhub workflow to simulate")



import re

import aiohttp
import yarl


async def token_authentication(api_token, verify_ssl=True):
    return aiohttp.ClientSession(
        headers={"Authorization": f"token {api_token}"},
        connector=aiohttp.TCPConnector(ssl=None if verify_ssl else False),
    )


async def basic_authentication(hub_url, username, password, verify_ssl=True):
    session = aiohttp.ClientSession(
        headers={"Referer": str(yarl.URL(hub_url) / "hub" / "api")},
        connector=aiohttp.TCPConnector(ssl=None if verify_ssl else False),
    )

    await session.post(
        yarl.URL(hub_url) / "hub" / "login",
        data={
            "username": username,
            "password": password,
        },
    )

    return session


async def keycloak_authentication(hub_url, username, password, verify_ssl=True):
    session = aiohttp.ClientSession(
        headers={"Referer": str(yarl.URL(hub_url) / "hub" / "api")},
        connector=aiohttp.TCPConnector(ssl=None if verify_ssl else False),
    )

    response = await session.get(yarl.URL(hub_url) / "hub" / "oauth_login")
    content = await response.content.read()
    auth_url = re.search('action="([^"]+)"', content.decode("utf8")).group(1)

    response = await session.post(
        auth_url.replace("&amp;", "&"),
        headers={"Content-Type": "application/x-www-form-urlencoded"},
        data={
            "username": username,
            "password": password,
            "credentialId": "",
        },
    )
    return session



from setuptools import setup, find_packages
import pathlib

here = pathlib.Path(__file__).parent.resolve()

long_description = (here / 'README.md').read_text(encoding='utf-8')


setup(
    name='jhub-client',
    version='0.1.7',
    description='Library and Client for managing, benchmarking, and interacting with jupyterhub',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/Quansight/jhub-client',
    author='Christopher Ostrouchov',
    author_email='chris.ostrouchov@gmail.com',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3 :: Only',
    ],
    keywords='jupyterhub, jupyter, benchmark',
    packages=find_packages(include=["jhub_client", "jhub_client.*"]),
    install_requires=['aiohttp', 'yarl'],
    extras_require={
        'dev': [
            'pytest',
            'pytest-asyncio',
            'black==22.3.0',
            'flake8==4.0.1',
        ],
    },
    entry_points={
        'console_scripts': [
            'jhubctl=jhub_client.__main__:main',
        ],
    },
)
